function ListTo( rsToList, nListDest, sDestFile, aFieldList, bNoRecnum, ;
                 sHeaderLine, sCaption, aTableProps )
   private dk_HeaderLine, dk_aFieldList, dk_HeaderType, dk_DestFile
   private dk_aTableProps
/* 
   ----------------------------------------------------------------
   Programmer......: David Kerber, Marko Mihorko
   Date............: 12/27/2001
   History.........: Version 1.0  12/08/2001
                     Version 1.1  12/08/2001 (added functionality 
                                              to generalize usage)
                     Version 1.1.1  12/31/2001 adds chooseprinter()
                     Version 1.1.2  1/12/2002 turns notificontrols
                                              off during printing
                     Version 1.1.2a  1/14/2002 Bug fix
                     Version 1.2  1/17/02 bug fixes, add custom 
                                          header line option
                     Version 1.2.3 1/18/02 More bug fixes, added the
                                           array option for the 
                                           header line.
                                              
   Modified by.....: David Kerber 12/08/2001
   
   Notes...........: This function emulates the XDML 'list' command 
                     for OODML rowsets, allowing you to do a quick 
                     listing of a rowset's contents to the screen, 
                     the printer, or anywhere else you can set, using 
                     the same options as the .output property of the 
                     report class, but without the overhead of a 
                     full-blown report in your project.  Optionally,
                     you can add many enhancements to your listing,
                     again with very little additional programming.

                     In all formats, you can add a caption line which
                     will print above the data table.  

                     All formats also have options for the header 
                     line.  If you specify no header line, then the
                     field names are used.  If you pass in an array, 
                     each array element becomes the header of a 
                     column.  If you pass in a string, that string is
                     printed verbatim, and you are responsible for 
                     the spacing and positioning of the words.

                     In html output, you cah have ListTo generate
                     either a complete html document, including 
                     everything from the <html> tag to the </html>
                     tag, or it can produce just the table itself,
                     allowing you to embed the table in a larger,
                     more complex html document.  To append the 
                     table to an already-existing file, just use the
                     'additive' keyword after the destination file 
                     name.  This will key ListTo to NOT produce the
                     <html> or <body> tag pairs.

                     In html output, the caption is automatically 
                     centered above or below the table, depending on
                     what attributes you assign to it.  The header 
                     entries are in table header <th> tags, which
                     centers each one over its respective column and 
                     makes it bold.  You can also add any allowed 
                     standard html tags to your table as described 
                     below in the aTableAttrs parameter.

                     The table attributes array can be either an
                     associative array (assocArray) or a simple one-
                     dimensional array.  See below for what proper-
                     ties go where in the array.

   Written for.....: Visual dBASE 7.x, dB2K
   
   Called by.......: Program
   
   Usage...........: ListTo( oRowset, nDestination, sDestFileName, 
                             aFieldList, bNoRecnum, sCustomHeaderLine,
                             sCaption )
   
   Example.........: ListTo( myRowset, 1, "", "" )
                        lists all fields of myrowset to printer
                     ListTo( myrowset, 0, "", "CompName, CompAddress, 
                             CompPhone", true )
                        lists specified fields to the results pane 
                        of the command window.without record numbers
                     ListTo( myrowse, 1, "", "", true, aFldList )
                        prints all fields, with column headers listed
                        in array aFldList.
                        
   Returns.........: 0 if successful
                     1 if field list is invalid

   Creates.........: If destination is a file, then creates speci-
                     fied file with output; otherwise nothing.
                     
   Parameters......: rsToList      = Rowset object to be listed out.

                     nListDest     = Number indicating destination; 
                                     same enumeration as .output
                                     property of the report object:
                                        0 = Command window results pane
                                        1 = Default printer
                                        2 = Printer file
                                        3 = Default
                                        4 = HTML file
                                        5 = CGI Response

                     sDestFile     = Filename for output option 2 or 4
                                     By using the keyword ' additive' 
                                     after the file name, it will ap-
                                     pend its output to an already-
                                     existing file.  Without this, it
                                     will overwrite any existing file.
                                     Also see html description for more
                                     information.

                     aFieldList    = Array containing fields to be 
                                     output.  Null for all fields.

                     bNoRecnum     = Boolean:  true means skip record
                                     numbers (equivalent to 'off' parm
                                     of 'list to' command)

                     sHeaderLine   = String or array containing custom 
                     	             header line to be printed instead 
                                     of standard field name list.  Can 
                                     be an empty string (for no header 
                                     line at all).  If null, default 
                                     header is used.
          
                                   	 For example, to use the column 
                                     headers from your grid:
                     	   
                                     aHeader = new Array ()
                                     for i = 1 to Form.Grid1.columns.size
                                        aHeader.add(Form.Grid1.columns[i]. ;
                                                headingControl.value)
                                     next i

                     sCaption      = String put out as a title bar or
                                     caption for the list.  In html, 
                                     can include formatting tags such
                                     as <h2>...</h2>, etc.

                     aTableProps   = Simple or AssocArray with tags 
                                     and settings for various sections 
                                     of the html table:  Note that 
                                     assocArray key names ARE CaSe
                                     SeNsItIvE and should be all UPPER
                                     CASE.  The tags used are the ones
                                     which were available in Netscape
                                     1.1.  HTML 4 now allows many other
                                     tags, which are not implemented
                                     here.  This is to ensure compati-
                                     bility with all browsers.

                        SIMPLE 
          ASSOCARRAY    ARRAY  
          KEY NAME      INDEX  DESCRIPTION

          BodyAttrs        1   Properties which go inside the
                               <body> tag

          BodyOpenTags     2   Tags which go immediately after
                               the <body> tag, such as <font>
                         
          BodyCloseTags    3   Tags before the </body> tag

          TableAttrs       4   Table properties which go inside the 
                               <table> tag, such as 
                               <table BGCOLOR="#EEEEEE">

          CaptionAttrs     5   Caption properties insice the
                               <caption> tag

          CaptionOpenTags  6   Tags after the <caption> tag

          CaptionCloseTags 7   Tags before the </caption> tag

          TRAttrs          8   Properties inside the <tr> tags.

          THAttrs          9   Properties inside the <th> tags.

          THOpenTags      10   Tags immediately after <th> tags.

          THCloseTags     11   Tags before </th> tags.

          TDAttrs         12   Properties inside the <td> tags.

          TDOpenTags      13   Tags immediately after <td> tags.

          TDCloseTags     14   Tags before </td> tags.


   Still to be done:  Finish implementing output options (CGI output).

   Suggested future enhancements:  Allow 'for' clause, as XDML 
                                       'list' command allows. 
                                   Return number of rows printed(?)
   ----------------------------------------------------------------
*/

   /* take rowset passed in rsToList, and list its data to the 
      destination indicated by nListDest, which uses the same 
      possible values as the .output property of the report class:  
           0=Window, 1=Printer, 2=Printer file, 3=Default, 
           4=HTML file, 5=CGI Response)
   */


   // store various settings which must be restored when we're done
   dk_bNotifyControls = rsToList.NotifyControls
   dk_sTalk = set( "talk" )
   dk_sPrint = set( "print" )
   dk_sConsole = set( "console" )
   set talk off

   dk_rtnParm = 0

   // First, set up the field list if it exists
   dk_aFieldList = aFieldList
   if type( "dk_aFieldList" ) <> "A"
         /* aFieldList does not exist as an array; create it from the
            rowset's fields
         */
         dk_afieldlist = new array( rstolist.fields.size )
         for dk_i = 1 to dk_afieldlist.size
            dk_afieldlist[ dk_i ] = rstolist.fields[ dk_i ].fieldname

            next dk_i
      endif
   dk_nFields = dk_aFieldList.size

   dk_DestFile = sDestFile

   if nListDest == 4 or nListDest == 5
      // if one of the html types
      /* build the Table Properties associative array if they passed in a
         simple array
      */
      dk_aTableProps = aTableProps
      if type( "dk_aTableProps" ) == "A"
         // simple array (assocArray's are type 'O')
         aTableProps = new AssocArray()
         aTableProps[ "BODYATTRS" ]        = iif( type( "dk_aTableProps[ 1 ]" ) ;
                                              == "L", "", dk_aTableProps[ 1 ] )
         aTableProps[ "BODYOPENTAGS" ]     = iif( type( "dk_aTableProps[ 2 ]" ) ;
                                              == "L", "", dk_aTableProps[ 2 ] )
         aTableProps[ "BODYCLOSETAGS" ]    = iif( type( "dk_aTableProps[ 3 ]" ) ;
                                              == "L", "", dk_aTableProps[ 3 ] )
         aTableProps[ "TABLEATTRS" ]       = iif( type( "dk_aTableProps[ 4 ]" ) ;
                                              == "L", "", dk_aTableProps[ 4 ] )
         aTableProps[ "CAPTIONATTRS" ]     = iif( type( "dk_aTableProps[ 5 ]" ) ;
                                              == "L", "", dk_aTableProps[ 5 ] )
         aTableProps[ "CAPTIONOPENTAGS" ]  = iif( type( "dk_aTableProps[ 6 ]" ) ;
                                              == "L", "", dk_aTableProps[ 6 ] )
         aTableProps[ "CAPTIONCLOSETAGS" ] = iif( type( "dk_aTableProps[ 7 ]" ) ;
                                              == "L", "", dk_aTableProps[ 7 ] )
         aTableProps[ "TRATTRS" ]          = iif( type( "dk_aTableProps[ 8 ]" ) ;
                                              == "L", "", dk_aTableProps[ 8 ] )
         aTableProps[ "THATTRS" ]          = iif( type( "dk_aTableProps[ 9 ]" ) ;
                                              == "L", "", dk_aTableProps[ 9 ] )
         aTableProps[ "THOPENTAGS" ]       = iif( type( "dk_aTableProps[ 10 ]" ) ;
                                              == "L", "", dk_aTableProps[ 10 ] )
         aTableProps[ "THCLOSETAGS" ]      = iif( type( "dk_aTableProps[ 11 ]" ) ;
                                              == "L", "", dk_aTableProps[ 11 ] )
         aTableProps[ "TDATTRS" ]          = iif( type( "dk_aTableProps[ 12 ]" ) ;
                                              == "L", "", dk_aTableProps[ 12 ] )
         aTableProps[ "TDOPENTAGS" ]       = iif( type( "dk_aTableProps[ 13 ]" ) ;
                                              == "L", "", dk_aTableProps[ 13 ] )
         aTableProps[ "TDCLOSETAGS" ]      = iif( type( "dk_aTableProps[ 14 ]" ) ;
                                              == "L", "", dk_aTableProps[ 14 ] )

      elseif type( "dk_aTableProps" ) = "L"
         // undefined
         aTableProps = new AssocArray( )
         aTableProps[ "BODYATTRS" ]        = ""
         aTableProps[ "BODYOPENTAGS" ]     = ""
         aTableProps[ "BODYCLOSETAGS" ]    = ""
         aTableProps[ "TABLEATTRS" ]       = ""
         aTableProps[ "CAPTIONATTRS" ]     = ""
         aTableProps[ "CAPTIONOPENTAGS" ]  = ""
         aTableProps[ "CAPTIONCLOSETAGS" ] = ""
         aTableProps[ "TRATTRS" ]          = ""
         aTableProps[ "THATTRS" ]          = ""
         aTableProps[ "THOPENTAGS" ]       = ""
         aTableProps[ "THCLOSETAGS" ]      = ""
         aTableProps[ "TDATTRS" ]          = ""
         aTableProps[ "TDOPENTAGS" ]       = ""
         aTableProps[ "TDCLOSETAGS" ]      = ""
         
      endif

   endif
   
   /* strip off the 'addi' clause if it exists, and put quotes around
      one version of the filename, since the 'set alte to' requires 
      quotes, and the file.open() cannot have them
   */
   temp = at( "ADDI", upper( dk_destfile ) )
   if temp <> 0 
         dk_destfile = left( dk_destfile, temp - 2 )
         dk_addi = "ADDI"
      else
         dk_addi = ""
      endif
   dk_destfile2 = "'" + dk_DestFile + "'"

   do case
      case nListDest == 0   // Window
         set console on

      case nListDest == 1   // Printer
         set console off
         rtnval = _app.printer.chooseprinter()
         if rtnval = false
               return
            endif
         set printer to
         set printer on

      case nListDest == 2   // Printer file
         set console off
         set alternate to &dk_DestFile2. &dk_addi
         set alternate on

      case nListDest == 3   // Default dest

      case nListDest == 4   // HTML file
         set console off

         // modified by JPMartel (start)
         if dk_addi == "ADDI"
            dk_file1 = new file()
            dk_file1.open( &dk_destfile2., 'RW' )
            dk_file1.seek( -32, 2 )   // move the pointer 17 char before the last character
            dk_file1.write( "<p><p>"+space(26) )  // delete those characters
            dk_file1.close()
         endif
         // modified by JPMartel (end)

         set alternate to &dk_destfile2. &dk_addi.
         set alternate on
         if dk_addi <> "ADDI"
               // not adding to an already-existing file
	            ?? '<!doctype html public "-//w3c//dtd html 4.0 transitional//en">'
	            ? "<html>"
               ? "<head></head>"
               ? buildtag( "body", atableprops, "BodyAttrs" )
               ? buildtag( "", atableprops, "BodyOpenTags" )
            endif
         ? buildtag( "table", atableprops, "TableAttrs" )
//         ? buildtag( "", atableprops, "TableOpenTags" )

      case nListDest == 5   // CGI output
         if dk_addi <> "ADDI"
               // not adding to an already-existing file
	            ? '<!doctype html public "-//w3c//dtd html 4.0 transitional//en">'
               ? "<html>"
               ? "<head></head>"
               ? buildtag( "body", atableprops, "BodyAttrs" )
               ? buildtag( "", atableprops, "BodyOpenTags" )
            endif
         ? buildtag( "table", atableprops, "TableAttrs" )
//         ? buildtag( "", atableprops, "TableOpenTags" )

      endcase

   rsToList.NotifyControls = false
   rsToList.first()

   /* put out the caption line */
   dk_Caption = sCaption
   if type( "dk_Caption" ) = "C"
         if nListDest = 4 or nListDest = 5
               // html or cgi
               dk_m = buildtag( "caption", atableprops, "CaptionAttrs" )
               dk_m += buildtag( "", atableprops, "CaptionOpenTags" )
               dk_m += sCaption
               dk_m += buildtag( "", atableprops, "CaptionCloseTags" )
               dk_m += "</caption>"
            else
               dk_m = sCaption
            endif
         ? dk_m
      endif

   /* do the header line:  either custom, or the fields list 
      as indicated by parameters passed in
   */
   dk_HeaderLine = sHeaderLine
   dk_HeaderType = type( "dk_HeaderLine" )
   if nlistdest = 4 or nlistdest = 5
         // html
         dk_m = buildtag( "tr", atableprops, "HeaderTRAttrs" )
//         dk_m += buildtag( "", atableprops, "HeaderTROpenTags" )
         dk_m += buildtag( "th", atableprops, "THAttrs" )
         dk_m += buildtag( "", atableprops, "THOpenTags" )
      else
         dk_m = ""
      endif

   do case
      case dk_headertype = "C"
         // use the custom header line as-is
         dk_m += dk_HeaderLine
         if nlistdest = 4 or nlistdest = 5
               // html
               dk_m = trim( dk_m )
               dk_m += buildtag( "", atableprops, "THCloseTags" )
               dk_m += "</th>"
            endif
         
      otherwise
         // use the default header line
         if .not. bNoRecnum
               // print the record number
               if dk_headertype = "A"
                     dk_m += dk_headerline[ 1 ]
                     dk_extra = 1
                  else
                     dk_m += 'Record#  '
                     dk_extra = 0
                  endif
            else
               // no record number ('off')
               dk_m += ' '
               dk_extra = 0
            endif

         if nlistdest = 4 or nlistdest = 5
               // html
               dk_m = trim( dk_m )
               dk_m += buildtag( "", atableprops, "THCloseTags" )
               dk_m += "</th>"
            endif

         for dk_j = 1 to dk_nFields
            for dk_i = 1 to rsToList.fields.size
               if upper( rstolist.fields[ dk_i ].fieldname ) == ;
                         upper( dk_aFieldList[ dk_j ] )
                     exit
                  endif
               next dk_i

               if dk_i > rsToList.fields.size
                     dk_rtnparm = 1
                     loop
                  endif

               if upper( rsToList.fields[ dk_i ].className ) # "LOCKFIELD"
                     if nlistdest = 4 or nlistdest = 5
                           dk_m += buildtag( "th", atableprops, "THAttrs" )
                           dk_m += buildtag( "", atableprops, "THOpenTags" )
                        endif
                  
                     if dk_headertype = "A"
                           dk_hd = dk_headerline[ dk_j + dk_extra ]
                        else
                           dk_hd = rsToList.fields[ dk_i ].fieldname
                        endif
                     
                     if nlistdest = 4 or nlistdest = 5
                           // html:  not justified
                           dk_m += dk_hd
                           dk_m += buildtag( "", atableprops, "THCloseTags" )
                           dk_m += "</th>"

                        elseif upper( rsToList.fields[ dk_i ].type ) $ ;
                             "CHARACTER LOGICAL DATE TIMESTAMP DATETIME"
                           // left-justify the field and fieldname
                           dk_m += dk_hd + space(pad(rstolist,true)) + " "
               
                        else
                           // right-justify the field and fieldname
                           dk_m += space(pad(rstolist,true)) + dk_hd

                        endif

                  endif

            next dk_j

      endcase
      
   ? dk_m   // print the header line string

   if nlistdest = 4 or nlistdest = 5
//         dk_m = buildtag( "", atableprops, "HeaderTRCloseTags" )
         ?? "</tr>"
      endif

   // now, start listing the data
   dk_n = 1
   do
      if nlistdest = 4 or nlistdest = 5
            dk_m = buildtag( "tr", atableprops, "TRAttrs" )
//            dk_m += buildtag( "", atableprops, "TROpenTags" )
            dk_m += buildtag( "td", atableprops, "TDAttrs" )
            dk_m += buildtag( "", atableprops, "TDOpenTags" )
         else
            dk_m = ""
         endif

      if .not. bNoRecNum
            dk_temp = str( dk_n, 7 ) + "  "
         else
            dk_temp = ' '
         endif

      if nlistdest = 4 or nlistdest = 5
            dk_m += ltrim( rtrim( dk_temp ) )
            dk_m += buildtag( "", atableprops, "TDCloseTags" )
            dk_m += "</td>"
         else
            dk_m += dk_temp
         endif


      for dk_j = 1 to dk_nfields
         // there was a field list, so find the correct field
         for dk_i = 1 to rstolist.fields.size
            if upper( rstolist.fields[ dk_i ].fieldname ) == ;
                    upper( dk_aFieldList[ dk_j ] )
                  exit
               endif

            next dk_i

            if dk_i > rstolist.fields.size
                  dk_rtnparm = 1
                  loop
               endif

         if rsToList.fields[ dk_i ].className # "LOCKFIELD"
               dk_temp = contents( rstolist )
               if dk_temp == null
                     dk_temp = ""
                     dk_temp2 = displaywidth( rstolist )

                  else
                     dk_temp2 = pad( rstolist,false )

                  endif

               if nlistdest = 4 or nlistdest = 5
                     dk_m += buildtag( "td", atableprops, "TDAttrs" )
                     dk_m += buildtag( "", atableprops, "TDOpenTags" )
                     dk_m += dk_temp
                     dk_m += buildtag( "", atableprops, "TDCloseTags" )
                     dk_m += "</td>"

                  else
                     dk_m += dk_temp + space( dk_temp2 ) + " "

                  endif

            endif

         next dk_j

      ? dk_m   // print the data line

      if nlistdest = 4 or nlistdest = 5
            dk_m = buildtag( "", atableprops, "TRCloseTags" )
            dk_m += "/<tr>"

         endif

      dk_n += 1
     
      until not rsToList.next()

      if nlistdest = 4 or nlistdest = 5
	    // finish the table and html
            dk_m = buildtag( "", atableprops, "TableCloseTags" )
            dk_m += "</table>"
            ?  dk_m
            * if dk_addi <> "ADDI"  // disabled by JP Martel
                  // were not adding to another file, so close it out
                  dk_m = buildtag( "", atableprops, "BodyCloseTags" )
                  dk_m += "</body>"
                  ?  dk_m
                  ?  "</html>"

            * endif                 // disabled by JP Martel

         endif

      // do the final cr/lf
      ?

      // if sending to html file, remove the ending EOF character
      close alternate
      if nlistdest = 4
***         if dk_addi <> "ADDI"
            dk_file1 = new file()
            dk_file1.open( dk_DestFile, 'RW' )
            dk_file1.seek( -1, 2 )   // move the pointer before the last character
            dk_char1 = dk_file1.read( 1 ) // read the next (i.e. last) character
            if dk_char1 == chr( 26 )
                  dk_file1.seek( -1 , 2 )  // move back to the last character
                  dk_file1.write( " " )  // overwrite the endOfFile character
               endif
            dk_file1.close()
***         endif
      endif

      // now restore various settings to their prior states
      close printer
      set print &dk_sPrint
      set talk &dk_sTalk.
      set console &dk_sConsole
      rsToList.NotifyControls = dk_bNotifyControls

return dk_rtnparm

function contents( rsToList )
   dk_fldtype = upper( rsToList.fields[ dk_i ].type )
   do case

      case dk_fldtype = "LOGICAL"
         contents = IIF( rsToList.fields[ dk_i ].value == true, ".T.", ".F." )

      case dk_fldtype = "NUMERIC" or ;
             dk_fldtype = "DOUBLE" or ;
             dk_fldtype = "FLOAT"
         contents = str(rsToList.fields[ dk_i ].value,;
                        rsToList.fields[ dk_i ].length,;
                        rsToList.fields[ dk_i ].decimalLength)

      case dk_fldtype = "LONG" or ;
             dk_fldtype = "AUTOINC"
         contents = str(rsToList.fields[ dk_i ].value,11)

      otherwise
         contents = rstolist.fields[ dk_i ].value
      
      endcase

return contents

function pad( rsToList, header )
   /* returns the number of spaces which need to be added to each displayed
      field to get the data lines to line up with the field names in the
      header line
   */
   dk_fldtype = upper( rsToList.fields[ dk_i ].type )
   if dk_headertype = "A"
         dk_fldnamelen = dk_HeaderLine[ dk_j ].length
      else
         dk_fldnamelen = rsToList.fields[ dk_i ].fieldname.length
      endif
   dk_dw = DisplayWidth( rsToList )
   do case
      case dk_fldtype == "CHARACTER"
         /* character fields can use the actual field length unless it 
            comes from a calculated field, in which case we need to use
            the data itself.  Since we can't tell whether or not a field
            is a calculated field, we need to do it this way always.
         */
         if header
               dk_pad = dk_dw - dk_fldnamelen
            else
               dk_pad = dk_dw - len( rstolist.fields[ dk_i ].value )
            endif

      case dk_fldtype == "LOGICAL"
         /* logical fields have a length of 1, but it takes 3 spaces to
            display their values, so we can't use the .length property
         */
         if header
               dk_pad = dk_dw - dk_fldnamelen
            else
               dk_pad = dk_dw - 3
            endif

      case dk_fldtype == "DATE"
         /* date fields always need a width of either 8 or 10 to display, 
            depending on whether set century is on or off, even though 
            the actual field size is always 8
         */
         dk_datewidth = 8 + iif( set( "CENTURY" ) = "ON", 2, 0 )
         if header
               dk_pad = dk_dw - dk_fldnamelen
            else
               dk_pad = dk_dw - dk_datewidth
            endif

      case dk_fldtype == "TIMESTAMP" or ;
               dk_fldtype == "DATETIME"
         /* timestamp fields always need a width of either 20 or 22 to display, 
            depending on whether set century is on or off, even though 
            the actual field size is always 8
         */
         dk_datewidth = 20 + iif( set( "CENTURY" ) = "ON", 2, 0 )
         if header
               dk_pad = dk_dw - dk_fldnamelen
            else
               dk_pad = dk_dw - dk_datewidth
            endif

      otherwise
         /* other fields use the actual field length */
         if header
               dk_pad = dk_dw - dk_fldnamelen
            else
               dk_pad = dk_dw - rsToList.fields[ dk_i ].length
            endif

      endcase

   return dk_pad

function DisplayWidth( rsToList )
   // determines the required width of the designated display field
   dk_fldtype = upper( rsToList.fields[ dk_i ].type )
   if dk_headertype = "A"
         dk_fldnamelen = dk_HeaderLine[ dk_j ].length
      else
         dk_fldnamelen = rsToList.fields[ dk_i ].fieldname.length
      endif
   do case
      case dk_fldtype == "CHARACTER"
         /* character fields can use the actual field length unless it 
            comes from a calculated field, in which case we need to use
            the data itself.  Since we can't tell whether or not a field
            is a calculated field, we need to do it this way always.
         */
         return max( dk_fldnamelen, len( rstolist.fields[ dk_i ].value ) )

      case dk_fldtype == "LOGICAL"
         /* logical fields have a length of 1, but it takes 3 spaces to
            display their values, so we can't use the .length property
         */
         return max( dk_fldnamelen, 3 )
      case dk_fldtype == "DATE"
         /* date fields always need a width of either 8 or 10 to display, 
            depending on whether set century is on or off, even though 
            the actual field size is always 8
         */
         datewidth = 8 + iif( set( "CENTURY" ) = "ON", 2, 0 )
         return max( dk_fldnamelen, datewidth )

      case dk_fldtype == "TIMESTAMP" or dk_fldtype == "DATETIME"
         /* timestamp fields always need a width of either 20 or 22 to display, 
            depending on whether set century is on or off, even though 
            the actual field size is always 8
         */
         datewidth = 20 + iif( set( "CENTURY" ) = "ON", 2, 0 )
         return max( dk_fldnamelen, datewidth )

      otherwise
         /* other fields use the actual field length */
         return max( dk_fldnamelen, rsToList.fields[ dk_i ].length )

      endcase

   return

function BuildTag( sTag, aTableProps, sPropKey )
   sBuiltTag = ""
   try
         if sTag == ""
               /* a blank sTag means these are additional tags before or
                  after a main table tag, not inside it
               */
               sBuiltTag =  aTableProps[ upper( sPropKey ) ]
            else
               /* a non-blank sTag means these are additional properties
                  which go inside the given tag
               */
               sBuiltTag = "<" + sTag + " " + aTableProps[ upper( sPropKey ) ] + ">"
            endif
      catch ( exception e )
         sBuiltTag = iif( sTag == "", "", "<" + sTag + ">" )
      endtry

   return sBuiltTag

